# mem-phys-addr.py: Resolve physical address samples
# SPDX-License-Identifier: GPL-2.0
#
# Copyright (c) 2018, Intel Corporation.

import os
import sys
import re
import bisect
import collections
from dataclasses import dataclass
from typing import (Dict, Optional)

sys.path.append(os.environ['PERF_EXEC_PATH'] + \
    '/scripts/python/Perf-Trace-Util/lib/Perf/Trace')

@dataclass(frozen=True)
class IomemEntry:
    """Read from a line in /proc/iomem"""
    begin: int
    end: int
    indent: int
    label: str

# Physical memory layout from /proc/iomem. Key is the indent and then
# a list of ranges.
iomem: Dict[int, list[IomemEntry]] = collections.defaultdict(list)
# Child nodes from the iomem parent.
children: Dict[IomemEntry, set[IomemEntry]] = collections.defaultdict(set)
# Maximum indent seen before an entry in the iomem file.
max_indent: int = 0
# Count for each range of memory.
load_mem_type_cnt: Dict[IomemEntry, int] = collections.Counter()
# Perf event name set from the first sample in the data.
event_name: Optional[str] = None

def parse_iomem():
    """Populate iomem from /proc/iomem file"""
    global iomem
    global max_indent
    global children
    with open('/proc/iomem', 'r', encoding='ascii') as f:
        for line in f:
            indent = 0
            while line[indent] == ' ':
                indent += 1
            if indent > max_indent:
                max_indent = indent
            m = re.split('-|:', line, 2)
            begin = int(m[0], 16)
            end = int(m[1], 16)
            label = m[2].strip()
            entry = IomemEntry(begin, end, indent, label)
            # Before adding entry, search for a parent node using its begin.
            if indent > 0:
                parent = find_memory_type(begin)
                assert parent, f"Given indent expected a parent for {label}"
                children[parent].add(entry)
            iomem[indent].append(entry)

def find_memory_type(phys_addr) -> Optional[IomemEntry]:
    """Search iomem for the range containing phys_addr with the maximum indent"""
    for i in range(max_indent, -1, -1):
        if i not in iomem:
            continue
        position = bisect.bisect_right(iomem[i], phys_addr,
                                       key=lambda entry: entry.begin)
        if position is None:
            continue
        iomem_entry = iomem[i][position-1]
        if  iomem_entry.begin <= phys_addr <= iomem_entry.end:
            return iomem_entry
    print(f"Didn't find {phys_addr}")
    return None

def print_memory_type():
    print(f"Event: {event_name}")
    print(f"{'Memory type':<40}  {'count':>10}  {'percentage':>10}")
    print(f"{'-' * 40:<40}  {'-' * 10:>10}  {'-' * 10:>10}")
    total = sum(load_mem_type_cnt.values())
    # Add count from children into the parent.
    for i in range(max_indent, -1, -1):
        if i not in iomem:
            continue
        for entry in iomem[i]:
            global children
            for child in children[entry]:
                if load_mem_type_cnt[child] > 0:
                    load_mem_type_cnt[entry] += load_mem_type_cnt[child]

    def print_entries(entries):
        """Print counts from parents down to their children"""
        global children
        for entry in sorted(entries,
                            key = lambda entry: load_mem_type_cnt[entry],
                            reverse = True):
            count = load_mem_type_cnt[entry]
            if count > 0:
                mem_type = ' ' * entry.indent + f"{entry.begin:x}-{entry.end:x} : {entry.label}"
                percent = 100 * count / total
                print(f"{mem_type:<40}  {count:>10}  {percent:>10.1f}")
                print_entries(children[entry])

    print_entries(iomem[0])

def trace_begin():
    parse_iomem()

def trace_end():
    print_memory_type()

def process_event(param_dict):
    if "sample" not in param_dict:
        return

    sample = param_dict["sample"]
    if "phys_addr" not in sample:
        return

    phys_addr  = sample["phys_addr"]
    entry = find_memory_type(phys_addr)
    if entry:
        load_mem_type_cnt[entry] += 1

    global event_name
    if event_name is None:
        event_name  = param_dict["ev_name"]



#!/usr/bin/env python3
# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
# Basic sanity check of perf JSON output as specified in the man page.

import argparse
import sys
import json

ap = argparse.ArgumentParser()
ap.add_argument('--no-args', action='store_true')
ap.add_argument('--interval', action='store_true')
ap.add_argument('--system-wide-no-aggr', action='store_true')
ap.add_argument('--system-wide', action='store_true')
ap.add_argument('--event', action='store_true')
ap.add_argument('--per-core', action='store_true')
ap.add_argument('--per-thread', action='store_true')
ap.add_argument('--per-cache', action='store_true')
ap.add_argument('--per-cluster', action='store_true')
ap.add_argument('--per-die', action='store_true')
ap.add_argument('--per-node', action='store_true')
ap.add_argument('--per-socket', action='store_true')
ap.add_argument('--metric-only', action='store_true')
ap.add_argument('--file', type=argparse.FileType('r'), default=sys.stdin)
args = ap.parse_args()

Lines = args.file.readlines()

def isfloat(num):
  try:
    float(num)
    return True
  except ValueError:
    return False


def isint(num):
  try:
    int(num)
    return True
  except ValueError:
    return False

def is_counter_value(num):
  return isfloat(num) or num == '<not counted>' or num == '<not supported>'

def check_json_output(expected_items):
  checks = {
      'counters': lambda x: isfloat(x),
      'core': lambda x: True,
      'counter-value': lambda x: is_counter_value(x),
      'cgroup': lambda x: True,
      'cpu': lambda x: isint(x),
      'cache': lambda x: True,
      'cluster': lambda x: True,
      'die': lambda x: True,
      'event': lambda x: True,
      'event-runtime': lambda x: isfloat(x),
      'interval': lambda x: isfloat(x),
      'metric-unit': lambda x: True,
      'metric-value': lambda x: isfloat(x),
      'metric-threshold': lambda x: x in ['unknown', 'good', 'less good', 'nearly bad', 'bad'],
      'metricgroup': lambda x: True,
      'node': lambda x: True,
      'pcnt-running': lambda x: isfloat(x),
      'socket': lambda x: True,
      'thread': lambda x: True,
      'unit': lambda x: True,
      'insn per cycle': lambda x: isfloat(x),
      'GHz': lambda x: True,  # FIXME: it seems unintended for --metric-only
  }
  input = '[\n' + ','.join(Lines) + '\n]'
  for item in json.loads(input):
    if expected_items != -1:
      count = len(item)
      if count not in expected_items and count >= 1 and count <= 7 and 'metric-value' in item:
        # Events that generate >1 metric may have isolated metric
        # values and possibly other prefixes like interval, core,
        # counters, or event-runtime/pcnt-running from multiplexing.
        pass
      elif count not in expected_items and count >= 1 and count <= 5 and 'metricgroup' in item:
        pass
      elif count - 1 in expected_items and 'metric-threshold' in item:
          pass
      elif count in expected_items and 'insn per cycle' in item:
          pass
      elif count not in expected_items:
        raise RuntimeError(f'wrong number of fields. counted {count} expected {expected_items}'
                           f' in \'{item}\'')
    for key, value in item.items():
      if key not in checks:
        raise RuntimeError(f'Unexpected key: key={key} value={value}')
      if not checks[key](value):
        raise RuntimeError(f'Check failed for: key={key} value={value}')


try:
  if args.no_args or args.system_wide or args.event:
    expected_items = [5, 7]
  elif args.interval or args.per_thread or args.system_wide_no_aggr:
    expected_items = [6, 8]
  elif args.per_core or args.per_socket or args.per_node or args.per_die or args.per_cluster or args.per_cache:
    expected_items = [7, 9]
  elif args.metric_only:
    expected_items = [1, 2]
  else:
    # If no option is specified, don't check the number of items.
    expected_items = -1
  check_json_output(expected_items)
except:
  print('Test failed for input:\n' + '\n'.join(Lines))
  raise



# perf script event handlers, generated by perf script -g python
# (c) 2010, Tom Zanussi <tzanussi@gmail.com>
# Licensed under the terms of the GNU GPL License version 2
#
# This script tests basic functionality such as flag and symbol
# strings, common_xxx() calls back into perf, begin, end, unhandled
# events, etc.  Basically, if this script runs successfully and
# displays expected results, Python scripting support should be ok.

from __future__ import print_function

import os
import sys

sys.path.append(os.environ['PERF_EXEC_PATH'] + \
	'/scripts/python/Perf-Trace-Util/lib/Perf/Trace')

from Core import *
from perf_trace_context import *

unhandled = autodict()

def trace_begin():
	print("trace_begin")
	pass

def trace_end():
	print_unhandled()

def irq__softirq_entry(event_name, context, common_cpu,
		       common_secs, common_nsecs, common_pid, common_comm,
		       common_callchain, vec):
	print_header(event_name, common_cpu, common_secs, common_nsecs,
		common_pid, common_comm)

	print_uncommon(context)

	print("vec=%s" % (symbol_str("irq__softirq_entry", "vec", vec)))

def kmem__kmalloc(event_name, context, common_cpu,
		  common_secs, common_nsecs, common_pid, common_comm,
		  common_callchain, call_site, ptr, bytes_req, bytes_alloc,
		  gfp_flags):
	print_header(event_name, common_cpu, common_secs, common_nsecs,
		common_pid, common_comm)

	print_uncommon(context)

	print("call_site=%u, ptr=%u, bytes_req=%u, "
		"bytes_alloc=%u, gfp_flags=%s" %
		(call_site, ptr, bytes_req, bytes_alloc,
		flag_str("kmem__kmalloc", "gfp_flags", gfp_flags)))

def trace_unhandled(event_name, context, event_fields_dict):
	try:
		unhandled[event_name] += 1
	except TypeError:
		unhandled[event_name] = 1

def print_header(event_name, cpu, secs, nsecs, pid, comm):
	print("%-20s %5u %05u.%09u %8u %-20s " %
		(event_name, cpu, secs, nsecs, pid, comm),
		end=' ')

# print trace fields not included in handler args
def print_uncommon(context):
	print("common_preempt_count=%d, common_flags=%s, "
		"common_lock_depth=%d, " %
		(common_pc(context), trace_flag_str(common_flags(context)),
		common_lock_depth(context)))

def print_unhandled():
	keys = unhandled.keys()
	if not keys:
		return

	print("\nunhandled events:\n")

	print("%-40s  %10s" % ("event", "count"))
	print("%-40s  %10s" % ("----------------------------------------",
				"-----------"))

	for event_name in keys:
		print("%-40s  %10d\n" % (event_name, unhandled[event_name]))
